#version 1.4

from vcCommand import *
import os
import csv
import xml.etree.ElementTree as et

cmd = getCommand()
app = cmd.Application

SIGNAL_LIST_HEADERS = [\
'line number',
'Symbol',
'Driver Signal Address',
'Type',
'default value',
'Comment',
'Definition internal Format',
'Definition external Format',
'userdef.column 1',
'userdef.column 2',
'userdef.column 3',
'userdef.column 4',
'userdef.column 5',
'macro name',
'standard',
'macro type',
'user 1',
'user 2',
'user 3',
'macro signal',
'X position of group',
'Y position of group',
'width of group',
'height of group',
'target file name',
'signal name',
'signal type',
'vc variable']

class template_signal_class(object):
  def __init__(self, header_line, signal_line, separator):
    self.separator = separator
    self.data = {}
    
    header_args = header_line.split(self.separator)
    signal_args = signal_line.split(self.separator)
    if len(header_args) != len(signal_args):
      return
    for i in range(len(header_args)):
      self.data[header_args[i]] = signal_args[i]
  
  
  def get_data_line(self, comp):
    values = []
    for column in SIGNAL_LIST_HEADERS:
      values.append(self.get_value(column, comp))
    data_line = self.separator.join(values)
    return data_line
  
  
  def add_connection_variable_element(self, comp, com_element, nsmap, parent_el):
    any_el = et.SubElement(parent_el, '{%s}anyType' % (nsmap['Array']))
    any_el.set('{%s}type' % (nsmap['i']), 'VariableGroupItem')
    
    display_el = et.SubElement(any_el, 'DisplayName')
    display_el.set('{%s}nil' % (nsmap['i']), 'true')
    
    server_el = et.SubElement(any_el, 'ServerItem')
    server_el.set('{%s}type' % (nsmap['i']), 'WinmodNet:WinmodVariableID')
    
    com_el = et.SubElement(server_el, '{%s}CommunicationElementName' % (nsmap['WinmodNet']))
    com_el.text = com_element
    
    dir_el = et.SubElement(server_el, '{%s}DataDirection' % (nsmap['WinmodNet']))
    if self.is_output():
      dir_el.text = 'Output'
    else:
      dir_el.text = 'Input'
    
    sig_el = et.SubElement(server_el, '{%s}SignalName' % (nsmap['WinmodNet']))
    sig_el.text = self.get_value('signal name', comp)
    
    sim_el = et.SubElement(any_el, 'SimulationVariable')
    
    comp_el = et.SubElement(sim_el, 'Component')
    comp_el.text = comp.Name
    
    beh = ''
    prop = ''
    args = self.get_value('vc variable', comp).split('.')
    if len(args) == 3:
      beh = args[1]
      prop = args[2]
    elif len(args) == 2:
      test_b = comp.findBehaviour(args[1])
      if test_b:
        beh = args[1]
      else:
        prop = args[1]
    
    beh_el = et.SubElement(sim_el, 'Behaviour')
    if beh:
      beh_el.text = beh
    else:
      beh_el.set('{%s}nil' % (nsmap['i']), 'true')
    
    prop_el = et.SubElement(sim_el, 'Property')
    if prop:
      prop_el.text = prop
    else:
      prop_el.set('{%s}nil' % (nsmap['i']), 'true')
  
  def get_value(self, column, comp = None):
    value = ''
    if column in self.data.keys():
      value = self.data[column]
      if comp:
        value = value.replace('COMP.', '%s.' % (comp.Name))
    return value
  
  def is_output(self):
    if self.get_value('signal type').lower() in ['bo', 'do', 'ao']:
      return True
    return False


class template_class(object):
  def __init__(self, template_file, separator):
    self.type = ''
    self.signals = []
    self.separator = separator
    
    if not os.path.exists(template_file):
      return
    head, tail = os.path.split(template_file)
    self.type = tail[:tail.find('.')]
    
    with open(template_file, 'r') as file:
      data = file.read()
    lines = data.split('\n')
    if len(lines) < 1:
      return
    header = lines[0]
    for i in range(1, len(lines)):
      data = lines[i]
      template_signal = template_signal_class(header, data, self.separator)
      if template_signal.get_value('signal name'):
        self.signals.append(template_signal)


  def get_data_lines(self, comp):
    lines = []
    for signal in self.signals:
      lines.append(signal.get_data_line(comp))
    return lines
  
  
  def add_connection_variables(self, comp, com_element, nsmap, sim_to_server_items_el, server_to_sim_items_el):
    for signal in self.signals:
      if signal.is_output():
        signal.add_connection_variable_element(comp, com_element, nsmap, sim_to_server_items_el)
      else:
        signal.add_connection_variable_element(comp, com_element, nsmap, server_to_sim_items_el)
    


def OnStart():
  #Entry point
  global prop_template, prop_only_selected, prop_separator, prop_hostname, prop_port, prop_com_element, prop_export
  
  prop_template = cmd.getProperty('Choose template by this property')
  prop_only_selected = cmd.getProperty('Export only selected components')
  prop_separator = cmd.getProperty('List separator character')
  prop_hostname = cmd.getProperty('WinMOD hostname')
  prop_port = cmd.getProperty('WinMOD port')
  prop_com_element = cmd.getProperty('WinMOD COM element path')
  prop_export = cmd.getProperty('Export')
  
  if not prop_template:
    prop_template = cmd.createProperty(VC_STRING, 'Choose template by this property')
    prop_template.Value = 'Category'
    
    prop_only_selected = cmd.createProperty(VC_BOOLEAN, 'Export only selected components')
    prop_only_selected.Value = False
    
    prop_separator = cmd.createProperty(VC_STRING, 'List separator character', VC_PROPERTY_STEP)
    prop_separator.StepValues = ['TAB', ';']
    prop_separator.Value = 'TAB'
    
    prop_hostname = cmd.createProperty(VC_STRING, 'WinMOD hostname')
    prop_hostname.Value = 'localhost'
    
    prop_port = cmd.createProperty(VC_INTEGER, 'WinMOD port')
    prop_port.Value = 40001
    
    prop_com_element = cmd.createProperty(VC_STRING, 'WinMOD COM element path')
    prop_com_element.Value = r'CoSim_Coupling\active layer\CoSim'
    
    prop_export = cmd.createProperty(VC_BUTTON, 'Export')
  
  prop_export.OnChanged = export
  app.clearMessages()
  executeInActionPanel()


def export(*args):
  # Export logic
  
  # get templates
  templates = get_templates()
  
  # get components
  if prop_only_selected.Value:
    comps = app.SelectionManager.getSelection(VC_SELECTION_COMPONENT)
  else:
    comps = app.Components
  
  comps.sort(key=(lambda x: x.Name))
  
  # Create files
  signal_file = write_signal_txt(templates, comps)
  if not signal_file:
    return
  connection_file = write_connection_xml(templates, comps, signal_file)
  if not connection_file:
    return
  
  print 'Created signal list file: ', signal_file
  print 'Created connection xml file: ', connection_file
  print 'Done.'


def write_signal_txt(templates, comps):
  # Creates signal list txt for WinMOD
  
  # form signal list
  lines = []
  lines.append(get_separator().join(SIGNAL_LIST_HEADERS))
  for comp in comps:
    template_prop = comp.getProperty(prop_template.Value)
    if not template_prop:
      continue
    type = str(template_prop.Value)
    if not type in templates.keys():
      continue
    template = templates[type]
    lines.extend(template.get_data_lines(comp))
  
  # open save dialog
  save_cmd = app.findCommand('dialogSaveFile')
  filter = 'text file | *.txt'
  title = 'Save signal list'
  save_cmd.Param_3 = filter
  save_cmd.Param_4 = title
  save_cmd.execute()
  if not save_cmd.Param_2:
    print 'Aborted.'
    return ''
  signal_file = save_cmd.Param_1[8:]
  signal_file = unicode(signal_file)
  with open(signal_file, 'w') as file:
    file.write('\n'.join(lines))
  
  return signal_file


def write_connection_xml(templates, comps, signal_file):
  # Creates connection xml file for VC
  
  # Load template xml
  cmd_uri = getCommandPath()[8:]
  cmd_folder, cmd_filename = os.path.split(cmd_uri)
  template_file = os.path.join(cmd_folder, 'connection_template.xml')
  if not os.path.exists(template_file):
    print r'Error. Cannot find connection xml template.'
    return ''
  with open(template_file) as file:
    template_data = file.read()
    
  # Form elementtree object
  nsmap = {}
  nsmap['System'] = r'http://www.w3.org/2001/XMLSchema'
  nsmap['Array'] = r'http://schemas.microsoft.com/2003/10/Serialization/Arrays'
  nsmap['ADS'] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.BeckhoffAds'
  nsmap['OpcUA'] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.OpcUA'
  nsmap['UrRtde'] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.UniversalRobotsRtde'
  nsmap['S7'] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.SiemensS7'
  nsmap['WinmodNet'] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.WinmodNet'
  nsmap['i'] = r'http://www.w3.org/2001/XMLSchema-instance'
  nsmap[''] = r'http://schemas.datacontract.org/2004/07/VisualComponents.Connectivity.Core'
  for key in nsmap.keys():
    et.register_namespace(key, nsmap[key])
  connection_xml = et.fromstring(template_data)
  settings_el = connection_xml.find('{%s}ConnectionPlugins/{%s}anyType/{%s}Servers/{%s}anyType/{%s}ConnectionSettings/.'  % (nsmap[''], nsmap['Array'], nsmap[''], nsmap['Array'], nsmap['']), nsmap)
  var_groups_el_el = connection_xml.find('{%s}ConnectionPlugins/{%s}anyType/{%s}Servers/{%s}anyType/{%s}VariableGroups/.'  % (nsmap[''], nsmap['Array'], nsmap[''], nsmap['Array'], nsmap['']), nsmap)
  sim_to_server_el = var_groups_el_el[0]
  sim_to_server_items_el = sim_to_server_el.find('{%s}Items' % nsmap[''], nsmap)
  server_to_sim_el = var_groups_el_el[1]
  server_to_sim_items_el = server_to_sim_el.find('{%s}Items' % nsmap[''], nsmap)
  
  # Settings
  settings_el.find('{%s}Hostname' % (nsmap['WinmodNet'])).text = prop_hostname.Value
  settings_el.find('{%s}Port' % (nsmap['WinmodNet'])).text = str(prop_port.Value)
  
  # Add signals
  for comp in comps:
    template_prop = comp.getProperty(prop_template.Value)
    if not template_prop:
      continue
    type = str(template_prop.Value)
    if not type in templates.keys():
      continue
    template = templates[type]
    template.add_connection_variables(comp, prop_com_element.Value, nsmap, sim_to_server_items_el, server_to_sim_items_el)
  
  # Write to file
  indent_etree(connection_xml)
  connection_file = signal_file[:signal_file.rfind('.')] + '.xml'
  with open(connection_file, 'w') as file:
    file.write(et.tostring(connection_xml, 'UTF-8'))
  
  return connection_file


def get_templates():
  # Load component templates
  
  cmd_uri = getCommandPath()[8:]
  cmd_folder, cmd_filename = os.path.split(cmd_uri)
  template_folder = os.path.join(cmd_folder, 'Templates')
  templates = {}
  
  for root, dirs, files in os.walk(template_folder):
    for file in files:
      template_file = os.path.join(root, file)
      template = template_class(template_file, get_separator())
      if template.type:
        templates[template.type] = template
  
  return templates


def indent_etree(elem, level = 0):
  # Indent element tree object
  one_indent = '  '
  i = '\n' + level * one_indent
  if len(elem):
    if not elem.text or not elem.text.strip():
      elem.text = i + one_indent
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
    for elem in elem:
      indent_etree(elem, level+1)
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
  else:
    if level and (not elem.tail or not elem.tail.strip()):
      elem.tail = i


def get_xml_element(root_el, element_path, index):
  # Find data xml element and return it.
  
  if root_el is None:
    return None
  
  args = element_path.split(r'/')
  parent_el = root_el
  while args:
    arg = args.pop(0)
    els = parent_el.findall(arg)
    print arg, parent_el, els
    get_index = 0
    if not args:
      get_index = index
    if len(els) <= get_index:
      return None
    el = els[get_index]
    parent_el = el
  
  return parent_el


def get_separator():
  if prop_separator.Value == 'TAB':
    return '\t'
  else:
    return prop_separator.Value